"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


import logging
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import colors
import numpy as np
from skimage import morphology, measure
import scipy.ndimage as ndi
import pandas as pd

from ctcontour.contours import ContourStep, Contour
from ctcontour.io_tools import load_dicom_image
from ctcontour.plot_tools import prep_figure, add_text_overlay, add_labelled_boundingbox, BoundingBox, draw_contour


class Slice:

    def __init__(self, fp):

        # Setup
        self.fp = fp
        self.ds = load_dicom_image(fp)
        self.rescaled_px = self._rescale()

        self.CTDIvol = self.ds.CTDIvol if 'CTDIvol' in self.ds else np.nan
        self.rows = self.ds.Rows if 'Rows' in self.ds else np.nan
        self.columns = self.ds.Columns if 'Columns' in self.ds else np.nan
        self.fov_mm = self.ds.ReconstructionDiameter if 'ReconstructionDiameter' in self.ds else np.nan

        self.has_contour = False
        self.contour = None
        self.ssde_mgy = np.nan
        self.phantom = np.nan

        # Truncation
        self.scan_limit_mask = np.nan

        self.out_of_scan_map = None  # map with out of scan pixels
        self.oos_px = np.nan  # count of how many px are out of scan
        self.is_out_of_scan = False  # bool whether patient in this slice is out of scan

        self.out_of_edge_map = None   # map with out of edge (l,r,t,b) pixels
        self.ooe_px = None  # tuple of (l,r,t,b) of counts of how many px are out of edge
        self.is_out_of_edge = False  # bool whether patient in this slice is out of edge

        self.is_truncated = False

        # Figures
        self.fig_detail = None
        self.fig_thumbs = None
        self.fig_truncation = None

    def _rescale(self):
        """
        Converts px data to HU, by applying rescale slope and intercept from DICOM headers
        :return: px data in Hounsfield Units
        """

        # Convert to int16 (from sometimes int16),
        # should be possible as values should always be low enough (<32k)
        try:
            px = self.ds.pixel_array.astype(np.int16)
            slope = self.ds.RescaleSlope
            intercept = self.ds.RescaleIntercept
        except AttributeError as e:
            logging.warning(f"{e}")
            return

        if slope != 1:
            px = slope * px.astype(np.float64)
            px = px.astype(np.int16)

        px += np.int16(intercept)

        return np.array(px, dtype=np.int16)

    def flag_truncation(self, small_objects_size=90, out_of_scan_tolerance=25, edge_tolerance=20, **kwargs):
        """
        There are two types of truncation:
        1. Out of scan. Patient is truncated because a portion lies beyound the reconstructed image
        2. Out of edge. Patient is truncated because a portion lies beyond either vertical or horizontal edges
        :param small_objects_size:
        :param out_of_scan_tolerance:
        :param edge_tolerance:
        :return:
        """

        # 1. Out of scan
        threshold = np.min(self.rescaled_px) + 1
        scan_limit_mask = np.where(self.rescaled_px <= threshold, True, False)
        scan_limit_mask = morphology.remove_small_objects(scan_limit_mask, small_objects_size)
        self.scan_limit_mask = morphology.dilation(scan_limit_mask, np.ones((4, 4)))

        self.out_of_scan_map = self.contour.contour & self.scan_limit_mask
        self.oos_px = np.count_nonzero(self.out_of_scan_map == True)

        if np.count_nonzero(self.out_of_scan_map == True) > out_of_scan_tolerance:
            self.is_out_of_scan = True
        else:
            self.is_out_of_scan = False

        # 2. Out of edge
        left = ((self.contour.contour[:, 0]) & (self.rescaled_px[:, 0] > threshold))
        right = ((self.contour.contour[:, -1]) & (self.rescaled_px[:, -1] > threshold))
        top = ((self.contour.contour[0]) & (self.rescaled_px[0] > threshold))
        bottom = ((self.contour.contour[-1]) & (self.rescaled_px[-1] > threshold))

        self.out_of_edge_map = np.full(self.rescaled_px.shape, False)
        self.out_of_edge_map[:, 0] = left
        self.out_of_edge_map[:, -1] = right
        self.out_of_edge_map[0] = top
        self.out_of_edge_map[-1] = bottom

        # tuple of counts of left, right, top, bottom
        self.ooe_px = (np.count_nonzero(self.out_of_edge_map[:, 0] == True),
                       np.count_nonzero(self.out_of_edge_map[:, -1] == True),
                       np.count_nonzero(self.out_of_edge_map[0] == True),
                       np.count_nonzero(self.out_of_edge_map[-1] == True))

        self.is_out_of_edge = False

        for edge in (left, right, top, bottom):
            if np.count_nonzero(edge == True) > edge_tolerance:
                self.is_out_of_edge = True

        if self.is_out_of_scan or self.is_out_of_edge:
            self.is_truncated = True

    def contour_with_ep(self, props):

        # Note on memory size:
        # Numpy uses int16 (16 bit) for pixel values
        # Numpy uses int8 (8 bit) for pixel bool values
        # For a 512x512px array:
        # 0.5MB for pixel values
        # 0.25MB for bool values
        # For this contour, approx 3.25MB of memory are needed to store all the steps

        # Initialise the contour steps
        steps = [ContourStep(f"Original", self.rescaled_px)]

        # Threshold
        px1 = self.rescaled_px > props['threshold']
        steps.append(ContourStep(f"Threshold: >{props['threshold']}HU", px1))

        # Despeckle
        px2 = morphology.remove_small_objects(px1, props['small_objects_size'])
        steps.append(ContourStep(f"Despeckle: <{props['small_objects_size']}px", px2))

        # Erode
        shape = (props['erosion_diam'], props['erosion_diam'])
        px3 = morphology.binary_erosion(px2, np.ones(shape))
        steps.append(ContourStep(f"Erode: {props['erosion_diam']}px", px3))

        # Despeckle
        px4 = morphology.remove_small_objects(px3, props['small_objects_size'])
        steps.append(ContourStep(f"Despeckle: <{props['small_objects_size']}px", px4))

        # Dilate
        shape = (props['dilation_diam'], props['dilation_diam'])
        px5 = morphology.binary_dilation(px4, np.ones(shape))
        steps.append(ContourStep(f"Dilate: {props['dilation_diam']}px", px5))

        # Eccentricity filter
        bb = BoundingBox(metric="eccentricity")  # show them before filtering
        steps.append(ContourStep(f"Contour eccentricities", px5, bb=bb))

        labels = measure.label(px5)  # Get bounding-box info for labelled regions
        selected_label_indexes = [r.label for r in measure.regionprops(labels) if r.eccentricity < props['eccentricity']]
        px6 = self._label_from_idx(labels, selected_label_indexes)
        steps.append(ContourStep(f"Filter eccentricities: >{props['eccentricity']}", px6))

        # Close
        px7 = morphology.binary_closing(px6, selem=morphology.disk(props['close_diam']))
        steps.append(ContourStep(f"Close: <{props['close_diam']}px", px7))

        # Binary filling
        px8 = ndi.binary_fill_holes(px7)
        steps.append(ContourStep(f"Fill", px8))

        #
        # Select n largest areas
        bb = BoundingBox(metric="area", number_format='0.0f')  # show them before filtering
        steps.append(ContourStep(f"Contour areas (px$^2$)", px8, bb=bb))
        #
        # get area labels
        labels = measure.label(px8)
        df = pd.DataFrame(measure.regionprops_table(labels, properties=['area']))
        df.index = np.arange(1, len(df) + 1)  # labels start from 1 (i.e. 0 == background)
        #
        # Sort and pick largest
        selected_label_indexes = list(df.sort_values('area', ascending=False).head(props['areas']).index)
        px9 = self._label_from_idx(labels, selected_label_indexes)
        steps.append(ContourStep(f"Final contour", px9))

        # Build Contour object
        self.contour = Contour(method='ep', props=props, steps=tuple(steps))
        self._build_contour_metrics(px9)

    def _label_from_idx(self, labels, selected_label_indexes):
        """
        returns a bool mask from the selected labels.
        :param selected_label_indexes: which label index (or indexes) to apply
        :param labels: labelled mask
        :return: px of same size as original, with zeros outside the mask
        """
        px_new = np.zeros_like(self.rescaled_px, dtype=np.int8)

        for lbl_idx in selected_label_indexes:
            px_new += np.where(labels == lbl_idx, 1, 0)

        px_new = np.where(px_new > 0, True, False)  # Flatten to a bool mask
        return px_new

    def _build_contour_metrics(self, px):
        """
        Use contour to perform measurements
        :param px:
        :return:
        """
        contour = measure.label(px).astype(bool)  # convert to Bool to flatten, otherwise contours won't find them all
        lbl = np.where(contour, 1, 0)
        df = pd.DataFrame(measure.regionprops_table(lbl, self.rescaled_px, properties=['area', 'mean_intensity']))

        # Add metrics
        self.contour.contour = contour
        self.contour.mpv = df.mean_intensity[0]  # Assume df has only 1 row
        self.contour.area_px2 = df.area[0]  # Assume df has only 1 row
        self.contour.area_mm2 = self._area_px_to_mm(self.contour.area_px2)
        self.contour.wed_cm2 = self.calc_wed()

        # Set flag
        self.has_contour = True

        # Calc SSDE
        self.ssde_mgy, self.phantom = self.calc_ssde()

    def calc_wed(self):
        """
        Equation for Water Equivalent Diameter
        :return: Water equivalent diameter in cm^2
        """
        return 2 * np.sqrt(
                             ((self.contour.mpv / 1000) + 1) *
                             (self.contour.area_mm2 / np.pi)
                           ) / 10  # div by 10 to convert to cm

    def calc_ssde(self):
        """
        Calculation of SSDE as per equations in AAPM220
        Requires these values to be present in DICOM tag:
        - CTDIvol
        - CTDIPhantomTypeCodeSequence
        :return: SSDE value in mGy
        """
        body, head = "IEC Body Dosimetry Phantom", "IEC Head Dosimetry Phantom"

        if 'CTDIPhantomTypeCodeSequence' in self.ds:
            code_meaning = self.ds.CTDIPhantomTypeCodeSequence[0].CodeMeaning
            if code_meaning == body:
                return self.CTDIvol * (3.704369 * np.exp(-0.03671937 * self.contour.wed_cm2)), body
            elif code_meaning == head:
                return self.CTDIvol * (1.874799 * np.exp(-0.03871313 * self.contour.wed_cm2)), head
            else:
                return np.nan, ''
        else:
            return np.nan, ''

    def _area_px_to_mm(self, area_px2):
        try:
            ps = self.ds.PixelSpacing
            return area_px2 * ps[0] * ps[1]  # Area to mm^2
        except KeyError as e:
            print(f'Key error:{e}')
            return np.nan

    def data_as_df(self):
        """
        Combine contour info as DataFrame with one row
        :return: DataFrame
        """
        return pd.DataFrame(
            dict(
                filepath=[self.fp],
                filename=[Path(self.fp).name],
                area_mm2=[self.contour.area_mm2],
                mpv_hu=[self.contour.mpv],
                wed_cm=[self.contour.wed_cm2],
                phantom=[self.phantom],
                ctdi_vol=[self.CTDIvol],
                ssde=[self.ssde_mgy],
                is_out_of_scan=self.is_out_of_scan,
                out_of_scan_px=self.oos_px,
                is_out_of_edge=self.is_out_of_edge,
                out_of_edge_px=str(self.ooe_px),
                is_truncated=self.is_truncated
            )
        )

    def draw_fig_detail(self):
        """
        Prepare a Figure to showcase all the contour steps.
        If a contour step includes a bounding box, show it.
        Final subplot will always draw the contour as an outline
        over the original pixel_array.
        :return: Figure
        """
        fig, axs = prep_figure(self.contour.steps)

        fig.suptitle(self.fp.name, fontsize=14, y=0.99)

        for step, ax in zip(self.contour.steps, axs):
            ax.imshow(step.pixel_array, cmap='gray')
            ax.set_title(step.label)

            if step.bb:  # If the contour step includes a bounding box
                labels = measure.label(step.pixel_array)
                df = pd.DataFrame(measure.regionprops_table(labels, properties=['bbox', step.bb.metric]))

                add_labelled_boundingbox(ax, df, step.bb.metric,
                                         number_format=step.bb.number_format,
                                         color=step.bb.color)

        # Show contour
        vmin = np.min(self.rescaled_px)
        vmax = np.max(self.rescaled_px)

        cutout = np.where(self.contour.contour, self.contour.steps[0].pixel_array, vmin)
        axs[-1].imshow(cutout, cmap='gray', vmin=vmin, vmax=vmax)

        self.fig_detail = fig
        return fig

    def draw_fig_thumbs(self):
        """
        Prepare a Figure showing only the original pixel array
        on the left and the contour overlay on right
        :return: Figure
        """
        # fig, axs = prep_figure(self.contour.steps, cols=2, rows=1, figheight=4, figwidth=7)
        plt.switch_backend('Agg')
        fig, axs = plt.subplots(nrows=1, ncols=2)
        fig.set(figheight=5, figwidth=9)

        for ax in axs:
            ax.axis('off')

        fig.suptitle(self.fp.name, fontsize=9, y=0.04)

        axs[0].imshow(self.contour.steps[0].pixel_array, cmap='gray')
        axs[0].set_title(self.contour.steps[0].label, fontsize=10)

        vmin = np.min(self.rescaled_px)
        vmax = np.max(self.rescaled_px)

        # cutout = np.where(self.contour.contour, self.contour.steps[0].pixel_array, vmin)
        # axs[1].imshow(cutout, cmap='gray', vmin=vmin, vmax=vmax)
        # axs[1].set_title(self.contour.steps[-1].label, fontsize=10)

        axs[1].imshow(self.rescaled_px, cmap='gray')
        draw_contour(axs[1], self.contour.contour, linewidth=1.0, color='w')

        fig.subplots_adjust(left=0.01, bottom=0, right=0.99, top=1, wspace=0.02, hspace=0)

        self.fig_thumbs = fig
        return fig

    def draw_fig_truncation(self, oos_color='red', ooe_color='orange', **kwargs):
        plt.switch_backend('Agg')

        fig, ax = plt.subplots()
        fig.set(figheight=8, figwidth=8)
        fig.set_tight_layout(True)
        ax.axis('off')

        cmap_slm = colors.ListedColormap(['orangered'])
        cmap_oos = colors.ListedColormap([oos_color])  # orange
        cmap_ooe = colors.ListedColormap([ooe_color])  # darkorange

        # Show original
        ax.imshow(self.rescaled_px, cmap='gray')
        ax.set_title(self.fp.name, fontsize=10)

        # Show scan limit
        # alpha = np.where(self.scan_limit_mask == True, 0.25, 0.)
        # ax.imshow(self.scan_limit_mask, alpha=alpha, cmap=cmap_slm)

        # Show out_of_scan_map
        alpha = np.where(self.out_of_scan_map == True, 1., 0.)
        ax.imshow(self.out_of_scan_map, alpha=alpha, cmap=cmap_oos)

        # Show out_of_edge_map
        ooe_presentation = morphology.dilation(self.out_of_edge_map, np.ones((6, 6)))
        alpha = np.where(ooe_presentation == True, 1., 0.)
        ax.imshow(ooe_presentation, alpha=alpha, cmap=cmap_ooe)

        # draw_contour(ax, self.contour.contour, linewidth=0.5, color='w')

        text_labels = [
            f"is_out_of_scan: {self.is_out_of_scan} {self.oos_px}",
            f"is_out_of_edge: {self.is_out_of_edge} l={self.ooe_px[0]}, r={self.ooe_px[1]}, t={self.ooe_px[2]}, b={self.ooe_px[3]}",
            f"is_truncated: {self.is_truncated}"
        ]
        # add_text_overlay(ax, text_labels)

        self.fig_truncation = fig

    def save_fig_thumbs(self, fp):
        self.fig_thumbs.savefig(fp)

    def save_fig_detail(self, fp):
        self.fig_detail.savefig(fp)

    def save_fig_trunk(self, fp):
        self.fig_truncation.savefig(fp)

    def save_csv(self, fp, index=False):
        df = self.data_as_df()
        df.to_csv(fp, index=index)

    def save_npz(self, fp):
        np.savez_compressed(fp, mask=self.contour.contour)




"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


import functools
import time
import logging


def timer(func):
    """
    Small tool to time functions
    :param func:
    :return:
    """
    # https://realpython.com/python-timer/#a-python-timer-decorator
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        tic = time.perf_counter()
        value = func(*args, **kwargs)
        toc = time.perf_counter()
        elapsed_time = toc - tic
        logging.info(f"Elapsed time: {elapsed_time:0.4f} seconds")
    return wrapper_timer



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


import logging
from glob import glob
from multiprocessing import Pool
from matplotlib import pyplot as plt
import pydicom

from ctcontour.io_tools import target_fp, merge_pdfs, merge_csvs, NotAxialImageError
from ctcontour.slice import Slice


def single_run(fp, morph_props, truncation_props, kwargs):
    """
    Portion of the overall contouring process that we may parallelise
    :param fp: file path to single DCM file
    :param morph_props: morph props to apply
    :param truncation_props: morph props for truncation detection
    :param kwargs: other options
    :return: None
    """

    # Load the file
    try:
        logging.info(f"Reading: {fp}")
        s = Slice(fp)
    except pydicom.errors.InvalidDicomError as e:
        return
    except NotAxialImageError as e:
        return

    # Contour
    if kwargs['method'] == "ep":
        s.contour_with_ep(morph_props)
        s.flag_truncation(**truncation_props)

    # Output
    if kwargs['detail']:
        s.draw_fig_detail()
        dst_fp = target_fp(kwargs['src_root'], kwargs['dst_root'], fp, kwargs['detail_stem'], kwargs['plot_filetype'])
        logging.info(f"Writing: {dst_fp}")
        s.save_fig_detail(dst_fp)
        plt.close(s.fig_detail)

    if kwargs['thumbs']:
        s.draw_fig_thumbs()
        dst_fp = target_fp(kwargs['src_root'], kwargs['dst_root'], fp, kwargs['thumbs_stem'], kwargs['plot_filetype'])
        logging.info(f"Writing: {dst_fp}")
        s.save_fig_thumbs(dst_fp)
        plt.close(s.fig_thumbs)

    if kwargs['trunk']:
        s.draw_fig_truncation(**truncation_props)
        dst_fp = target_fp(kwargs['src_root'], kwargs['dst_root'], fp, kwargs['trunk_stem'], kwargs['plot_filetype'])
        logging.info(f"Writing: {dst_fp}")
        s.save_fig_trunk(dst_fp)

    if kwargs['csv']:
        dst_fp = target_fp(kwargs['src_root'], kwargs['dst_root'], fp, '', 'csv')
        logging.info(f"Writing: {dst_fp}")
        s.save_csv(dst_fp)

    if kwargs['npz']:
        dst_fp = target_fp(kwargs['src_root'], kwargs['dst_root'], fp, '', 'npz')
        logging.info(f"Writing: {dst_fp}")
        s.save_npz(dst_fp)


def starmap_run(fp, morph_props, truncation_props, kwargs):
    """
    Unravel processing on all available cores
    :param fp: file path to single DCM file
    :param morph_props: morph props to apply
    :param kwargs: other options
    :return: None
    """
    single_run(fp, morph_props, truncation_props, kwargs)


def merge_files(kwargs):
    """
    Outside merge function to prepare list of files for merging.
    This is needed to iterate through the multiple nested dirs
    :param kwargs: Dict with opts defined in config.py
    :return: None
    """
    dst_dirs = [kwargs['dst_root']] + [fp for fp in kwargs['dst_root'].rglob('*') if fp.is_dir()]

    if kwargs['merge_pdf']:
        for dst_dir in dst_dirs:

            def _merge_pdf(suffix):
                logging.info(f"Merging [{kwargs[suffix]}] PDFs in {dst_dir}")
                merge_pdfs(
                    sorted([f for f in glob(str(dst_dir) + f"/*{kwargs[suffix]}.pdf")]),
                    dst_dir / f"{dst_dir.name}{kwargs[suffix]}.pdf"
                )

            if kwargs['thumbs']:
                _merge_pdf('thumbs_stem')

            if kwargs['detail']:
                _merge_pdf('detail_stem')

            if kwargs['trunk']:
                _merge_pdf('trunk_stem')

    if kwargs['merge_csv']:
        for dst_dir in dst_dirs:
            logging.info(f"Merging CSVs in {dst_dir}")
            merge_csvs(
                sorted([f for f in glob(str(dst_dir) + f"/*.csv")]),
                dst_dir / f"{dst_dir.name}.csv"
            )


def start(morph_props=None, truncation_props=None, kwargs=None):
    """
    This is the overall contouring process. Here we perform:
    - Build list of directories
    - Contour all valid files in each directory
    - Merge PDF and CSV results
    :param morph_props: Dict with props. Must match contour_method in kwargs
    :param truncation_props: morph props for truncation detection
    :param kwargs: Dict with options defined in config.py
    :return: None
    """

    # If it's a single file, process and return
    if kwargs['src_root'].is_file():
        single_run(kwargs['src_root'], morph_props, truncation_props, kwargs)
        return

    # if it's a folder, check for recursion
    if kwargs['recursive']:
        files = sorted([fp for fp in kwargs['src_root'].rglob('*') if fp.is_file() and fp.suffix == '.dcm'])
        dirs = [kwargs['src_root']] + [fp for fp in kwargs['src_root'].rglob('*') if fp.is_dir()]
        logging.info(f"Found {len(files)} files in {len(dirs)} folders")
    else:
        files = sorted([fp for fp in kwargs['src_root'].glob('*') if fp.is_file() and fp.suffix == '.dcm'])
        logging.info(f"Found {len(files)} files")

    # Process the files
    if kwargs['single']:
        for fp in files:
            single_run(fp, morph_props, truncation_props, kwargs)
    else:
        pool_kwargs = [(fp, morph_props, truncation_props, kwargs) for fp in files]
        pool = Pool()
        pool.starmap(starmap_run, pool_kwargs)

    # Merging
    merge_files(kwargs)



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


import argparse
import logging.config
from pathlib import Path

from ctcontour import contour_proc
from ctcontour.config import MORPH_PROPS_EP, TRUNCATION_PROPS, OPTIONS
from ctcontour.decorator_funcs import timer

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s [%(levelname)s] - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')


def parse_args_contour():
    message = """
    Fully automated patient abdomino-pelvic contouring for CT images.
    https://doi.org/10.1016/j.ejmp.2022.10.027.
    """

    parser = argparse.ArgumentParser(description=message)

    parser.add_argument('src_root', type=str, metavar="Source",
                        help='Location of source dicom file or folder to contour')

    parser.add_argument('dst_root', type=str, metavar="Destination",
                        help='Destination for results. Should ideally be an empty directory. Existing files may be overwritten \n')

    parser.add_argument("--thumbs",
                        help="Save PDF showing original image at left and contoured image at right. \
                              File saved as *_thumbs.pdf",
                        action="store_true")

    parser.add_argument("--detail",
                        help="save PDF showing full morphological steps. \
                        File saved as *_detail.pdf",
                        action="store_true")

    parser.add_argument("--trunk",
                        help="save PDF showing truncation map. Red for out-of-scan and orange for out-of-edge. \
                        File saved as *_trunk.pdf",
                        action="store_true")

    parser.add_argument("--npz",
                        help="save trunk contour to compressed npz file. Uses dictionary key: 'mask': mask. \
                              File saved as *.npz",
                        action="store_true")

    parser.add_argument("--csv",
                        help="save contour metrics to csv. \
                              File saved as *.csv",
                        action="store_true")

    parser.add_argument("--single",
                        help="No multicore processing (useful for debugging)",
                        action="store_true")

    parser.add_argument("--recursive",
                        help="Also process subfolders. Generates a list of all subfolders before starting.",
                        action="store_true")

    parser.add_argument("--merge_pdf",
                        help="Merge generated PDFs. \
                              By default merges into [foldername]_[thumbs|detail|trunk].pdf",
                        action="store_true")

    parser.add_argument("--merge_csv",
                        help="Merge generated CSVs. \
                              By default merges into [foldername]_mask.csv",
                        action="store_true")

    args = parser.parse_args()

    args.src_root = Path(args.src_root)
    args.dst_root = Path(args.dst_root)

    return args


@timer
def contour():
    """
    Top level entry point for contouring and truncation detection.
    :return: None
    """
    args = parse_args_contour()

    kwargs = {**vars(args), **OPTIONS}

    logging.info(
         f"Running...\n"
         f"     Source: {kwargs['src_root']}\n"
         f"Destination: {kwargs['dst_root']}\n"
         f"  Recursive: {kwargs['recursive']}\n"
         f"Single core: {kwargs['single']}\n"
         f"   Save PDF: Thumbnails [{kwargs['thumbs']}], Details [{kwargs['detail']}], Trunk [{kwargs['trunk']}]\n"
         f"   Save CSV: {kwargs['csv']}\n"
         f"   Save NPZ: {kwargs['npz']}\n"
         f"      Merge: PDF [{kwargs['merge_pdf']}], CSV [{kwargs['merge_csv']}]\n")

    contour_proc.start(morph_props=MORPH_PROPS_EP, truncation_props=TRUNCATION_PROPS, kwargs=kwargs)


def __main__():
    contour()


if __name__ == "__main__":
    __main__()



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


from dataclasses import dataclass
import numpy as np
import pandas as pd

from ctcontour.plot_tools import BoundingBox


@dataclass
class ContourStep:
    label: str
    pixel_array: np.ndarray
    bb: BoundingBox = None


@dataclass
class Contour:
    method: str
    props: dict
    steps: tuple = None
    contour: np.ndarray = None
    area_px2: float = np.nan
    area_mm2: float = np.nan
    mpv: float = np.nan
    wed_cm2: float = np.nan

    def get_metrics(self, as_type=None):
        data = {
            "contour_method": [self.method],
            "area_px2": [self.area_px2],
            "area_mm2": [self.area_mm2],
            "mpv": [self.mpv],
            "wed_cm2": [self.wed_cm2]
        }
        if as_type == 'dict':
            return data
        else:
            return pd.DataFrame(data=data)



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


import os
import logging
from pathlib import Path

import pydicom
from PyPDF2 import PdfFileMerger, PdfFileReader
from PyPDF2 import utils as PyPDF2utils
import pandas as pd

from ctcontour.config import DCM_TAGS


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class NotAxialImageError(Error):
    """Not an Axial image (possibly a localiser or secondary image"""
    def __init__(self, image_type, message="Not an Axial image"):
        self.image_type = image_type
        self.message = f"{message}: {self.image_type}"
        super().__init__(self.message)


def load_dicom_image(fp):
    """
    Load ds from DICOM files

    :param fp: path to single DCM
    :return: Pydicom DataSet
    """

    try:
        ds = pydicom.dcmread(fp, specific_tags=DCM_TAGS)

        if 'AXIAL' not in list(ds.ImageType):
            raise NotAxialImageError(list(ds.ImageType))
        else:
            return ds

    except pydicom.errors.InvalidDicomError as e:
        logging.error(f"{fp}: {e}")
        raise
    except AttributeError as e:
        logging.error(f"{fp}: {e}")
        raise
    except NotAxialImageError as e:
        logging.error(f"{fp}: {e}")
        raise


def _delete_file(fp):
    try:
        os.remove(fp)
    except FileNotFoundError as e:
        logging.error(f"{fp} does not exist")
    except IsADirectoryError as e:
        logging.error(f"{fp} is a directory")


def target_fp(src_root, dst_root, src, stem_part, suffix):
    """
    Builds final target file path for writing
    :param src_root: top level source path
    :param dst_root: top level destination path
    :param src: fully qualified path to source dcm file
    :param stem_part: to be added at the end of the filename, like '_thumbs'
    :param suffix: the suffix, like pdf or csv
    :return: destination file path as Path object
    """

    # Convert to Paths if not done already
    src_root = Path(src_root)

    # If src_root points to a file, then program is running on a single file only
    if src_root.is_file():
        src_root = src_root.parent

    dst_root = Path(dst_root)
    src = Path(src)

    # Build target
    index = len(src_root.parts)
    dst = dst_root.joinpath(*src.parts[index:])
    dst = dst.parent / f"{dst.stem}{stem_part}.{suffix}"

    # Create folders if they don't exist
    if not dst.parent.exists():
        try:
            os.makedirs(dst.parent)
        except FileExistsError as e:
            logging.warning(f"Exists: {dst}")
    # if not dst.parent.exists():
    #
    # if dst.exists():
    #     logging.warning(f"File exists: {dst}")

    return dst


def merge_pdfs(pdf_list, fp, chunksize=500, delete_original=True):
    """
    Will merge a given list of pdf filepaths
    :param pdf_list: list of pdf filepaths
    :param fp: destination file path for merged result
    :param chunksize: Generate multiple files if list larger than chunksize
    :param delete_original: optionally remove the original list of pdf files
    :return: None
    """
    fp = Path(fp)
    chunks = [pdf_list[x:x + chunksize] for x in range(0, len(pdf_list), chunksize)]

    for chunknum, pdf_sublist in enumerate(chunks):
        merger = PdfFileMerger()
        for pdf in pdf_sublist:
            try:
                merger.append(PdfFileReader(open(pdf, 'rb')))
            except PyPDF2utils.PdfReadError as e:
                logging.warning(e)

        dst = fp.parent / f"{fp.stem}_{chunknum}{fp.suffix}"
        merger.write(str(dst))
        merger.close()

    if delete_original:
        for pdf in pdf_list:
            _delete_file(pdf)


def merge_csvs(csv_list, fp, index=False, delete_original=True):
    """
    Will merge a given list of csv filepaths
    :param csv_list: list of csv filepaths
    :param fp: destination file path for merged result
    :param index: Include Pandas index in output?
    :param delete_original: optionally remove the original list of pdf files
    :return: None
    """

    if not csv_list:
        return

    # Remove any existing target csv files otherwise we'll append old results with new ones
    try:
        if fp.exists():
            csv_list.remove(str(fp))
    except ValueError as e:
        logging.warning(f"Merging csv: {e}")

    # combine results
    df = pd.concat((pd.read_csv(f) for f in csv_list))
    df.to_csv(fp, index=index)

    # delete individual files
    if delete_original:
        for csv in csv_list:
            _delete_file(csv)



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""

import setuptools

""" CTcontour setup """


with open('README.md') as f:
    README = f.read()

setuptools.setup(
    author="Eric Pace",
    author_email="ericpace@pm.me",
    name='ctcontour',
    license="GNU GPLv3",
    description='Fully automated patient abdomino-pelvic contouring for CT images. https://doi.org/10.1016/j.ejmp.2022.10.027.',
    version='v1.1',
    long_description='Fully automated patient abdomino-pelvic contouring for CT images. https://doi.org/10.1016/j.ejmp.2022.10.027.',
    url='https://github.com/ericpace/ct_contour',
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'ctc=ctcontour.__main__:contour',
        ]
    },
    python_requires=">=3.8",
    # Enable install requires when publishing on the normal PyPi
    install_requires=[
        'pydicom==2.1.2',
        'numba==0.52.0',
        'scipy==1.4.1',
        'scikit-image==0.18.2',
        'pandas==1.2.1',
        'matplotlib==3.3.4',
        'seaborn==0.11.1',
        'PyPDF2==1.26.0'
    ],
    classifiers=[
        'Programming Language :: Python'
    ],
)



"""
Author: Eric Pace
This file is part of CTContour.

CTContour is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation version 3.

CTContour is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Patient CT Contour.
If not, see <https://www.gnu.org/licenses/>.
"""


from dataclasses import dataclass
from math import ceil
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from skimage import measure


@dataclass
class BoundingBox:
    metric: str = None
    number_format: str = '0.3f'
    color: str = 'w'


def prep_figure(contour_steps, cols=4, rows=None, figheight=14, figwidth=17):
    """
    Prepares a standard matplotlib subplots layout
    :param contour_steps: number of steps in the contour process
    :param cols: predefined number of columns
    :param rows: predefined number of rows
    :param figheight: desired figure height
    :param figwidth:  desired figure width
    :return: figure and axes objects
    """
    plt.switch_backend('Agg')

    if not rows:
        rows = ceil(len(contour_steps)/cols)

    fig, axes = plt.subplots(nrows=rows, ncols=cols)

    axs = axes.flatten()

    fig.set(figheight=figheight, figwidth=figwidth)
    fig.set_tight_layout(True)

    for ax in axs:
        ax.axis('off')
    return fig, axs


def add_labelled_boundingbox(ax, prop_df, property, number_format='0.2f', color='r'):
    """
    Overlays bounding boxes around label objects
    :param ax: input axis figure to overlay
    :param prop_df: DataFrame with property labels
    :param property: shape property label (e.g. area or eccentricity)
    :param number_format: how to format the number
    :param color: label colour
    :return: None
    """

    for index, roi in prop_df.iterrows():
        width = roi['bbox-3'] - roi['bbox-1']
        height = roi['bbox-2'] - roi['bbox-0']
        info_label = roi[property]

        boundingbox = Rectangle((roi['bbox-1'], roi['bbox-0']), width, height,
                                linewidth=0.5, edgecolor=color, facecolor='none')
        ax.add_patch(boundingbox)

        plot_size = ax.get_xlim()[1] - ax.get_xlim()[0]

        # Aligns label left or right depending on where bounding box is
        if roi['bbox-1'] <= plot_size/2:
            ax.text(roi['bbox-1'], roi['bbox-0'] - 2,
                    f"{info_label:{number_format}}", fontsize=8, color=color)
        else:
            ax.text(roi['bbox-3'], roi['bbox-0'] - 2,
                    f"{info_label:{number_format}}", fontsize=8, color=color, horizontalalignment='right')


def add_text_overlay(ax, text_str_list, color='w'):
    step = 15
    # top left alignment
    for i, text_str in enumerate(text_str_list):
        ax.text(5, 15+(i*step), text_str, fontsize=8, color=color)

    return ax


def draw_contour(ax, px, **kwargs):
    """
    Draw the contour outline
    :param ax: the Axes to plot on
    :param px: the bool mask with the contour
    :param kwargs: plotting options compatible with ax.plot(), like linewidth and color
    :return: None
    """
    contours = measure.find_contours(px)

    for contour in contours:
        ax.plot(contour[:, 1], contour[:, 0], **kwargs)